@@ -4,7 +4,7 @@ package test;
 
 =head1 NAME
 
-test - tools for helping in test suites (not including running external programs).
+test - tools for helping in test suites (not including running externalprograms).
 
 =head1 SYNOPSIS
 
@@ -46,6 +46,8 @@ Setting up the environment includes:
 
 =over 4
 
+=item Prepending F<blib/script> onto the path
+
 =item Pushing the module F<lib/> dir onto the @PERL5LIB var
 
 For executed scripts.
@@ -109,19 +111,23 @@ The following symbols are exported upon request:
 
 =item tempdir
 
+=item find_exec
+
+=item read_file
+
 =back
 
 =cut
 
 @EXPORT_OK = qw( BIN_DIR DATA_DIR REF_DIR LIB_DIR
-                 check_req compare evcheck only_files
+                 check_req compare evcheck find_exec only_files read_file
                  save_output restore_output tempdir tmpnam );
 
 # Utility -----------------------------
 
 use Carp                          qw( carp croak );
 use Cwd                      2.01 qw( cwd );
-use Env                           qw( PERL5LIB );
+use Env                           qw( PATH PERL5LIB );
 use Fatal                    1.02 qw( close open seek sysopen unlink );
 use Fcntl                    1.03 qw( :DEFAULT );
 use File::Compare          1.1002 qw( );
@@ -153,6 +159,30 @@ sub updir {
   File::Spec->updir(@_);
 }
 
+sub min {
+  croak "Can't min over 0 args!\n"
+    unless @_;
+  my $min = $_[0];
+  for (@_[1..$#_]) {
+    $min = $_
+      if $_ < $min;
+  }
+
+  return $min;
+}
+
+sub max {
+  croak "Can't max over 0 args!\n"
+    unless @_;
+  my $max = $_[0];
+  for (@_[1..$#_]) {
+    $max = $_
+      if $_ > $max;
+  }
+
+  return $max;
+}
+
 # -------------------------------------
 # PACKAGE CONSTANTS
 # -------------------------------------
@@ -162,15 +192,19 @@ use constant DATA_DIR => catdir $Bin, updir, 'data';
 use constant REF_DIR  => catdir $Bin, updir, 'testref';
 use constant LIB_DIR  => catdir $Bin, updir, 'lib';
 
+use constant BUILD_SCRIPT_DIR => => catdir $Bin, updir, qw( blib script );
+
 # -------------------------------------
 # PACKAGE ACTIONS
 # -------------------------------------
 
 # @PERL5LIB not available in Env for perl 5.00503
 # unshift @PERL5LIB, LIB_DIR;
-$PERL5LIB = join ':', LIB_DIR, $PERL5LIB;
+$PERL5LIB = defined $PERL5LIB ? join(':', LIB_DIR, $PERL5LIB) : LIB_DIR;
 unshift @INC,      LIB_DIR;
 
+$PATH = join ':', BUILD_SCRIPT_DIR, split /:/, $PATH;
+
 $_ = rel2abs($_)
   for @INC;
 
@@ -340,12 +374,16 @@ sub save_output {
     unless @_ == 2;
   my ($name, $filehandle) = @_;
 
+  my $tmpfh  = do { local *F; *F; };
+  my $savefh = do { local *F; *F; };
+
   my $tmpnam = POSIX::tmpnam;
-  sysopen my $tmpfh, $tmpnam, O_RDWR | O_CREAT | O_EXCL;
+  sysopen $tmpfh, $tmpnam, O_RDWR | O_CREAT | O_EXCL;
   unlink $tmpnam;
   select((select($tmpfh), $| = 1)[0]);
 
-  open my $savefh, '>&' . fileno $filehandle
+
+  open $savefh, '>&' . fileno $filehandle
     or die "can't dup $name: $!";
   open $filehandle, '>&' . fileno $tmpfh
     or die "can't open $name to tempfile: $!";
@@ -453,20 +491,33 @@ context.
 =cut
 
 my @tmpfns;
-sub tmpnam {
-  my $tmpnam = POSIX::tmpnam;
 
-  if (@_) {
-    push @tmpfns, [ $tmpnam, $_[0] ];
-  } else {
-    push @tmpfns, $tmpnam;
-  }
+BEGIN {
+  my $savewarn = $SIG{__WARN__};
+  # Subvert bizarre (& incorrect) subroutine redefined errors in 5.005_03
+  local $SIG{__WARN__} =
+    sub {
+      $savewarn->(@_)
+        if defined $savewarn                        and
+           UNIVERSAL::isa($savewarn,'CODE')         and
+           $_[0] !~ /^Subroutine tmpnam redefined/;
+    };
 
-  if (wantarray) {
-    sysopen my $tmpfh, $tmpnam, O_RDWR | O_CREAT | O_EXCL;
-    return $tmpnam, $tmpfh;
-  } else {
-    return $tmpnam;
+  *tmpnam = sub {
+    my $tmpnam = POSIX::tmpnam;
+
+    if (@_) {
+      push @tmpfns, [ $tmpnam, $_[0] ];
+    } else {
+      push @tmpfns, $tmpnam;
+    }
+
+    if (wantarray) {
+      sysopen my $tmpfh, $tmpnam, O_RDWR | O_CREAT | O_EXCL;
+      return $tmpnam, $tmpfh;
+    } else {
+      return $tmpnam;
+    }
   }
 }
 
@@ -533,12 +584,16 @@ END {
       if ( defined $ENV{TEST_DEBUG} and $ENV{TEST_DEBUG} =~ /\bSAVE\b/ ) {
         printf "Used temp dir: %s (%s)\n", @$_;
       } else {
+        # Solaris gets narky about removing the pwd.
+        chdir File::Spec->rootdir;
         rmtree $_->[0];
       }
     } else {
       if ( defined $ENV{TEST_DEBUG} and $ENV{TEST_DEBUG} =~ /\bSAVE\b/ ) {
         print "Used temp dir: $_\n";
       } else {
+        # Solaris gets narky about removing the pwd.
+        chdir File::Spec->rootdir;
         rmtree $_;
       }
     }
@@ -549,96 +604,340 @@ END {
 
 =head2 compare
 
-  ok compare($fn1, $fn2), 1, 'test name';
+  compare(+{ fn1 => $fn1, fn2 => $fn2, gzip => 1 });
 
-Compare two files, return 1 if they are the same, 0 if they differ, -1 if file
-cannot be read, -2 if file 2 cannot be read, -4 if both cannot be read.  -8
-for a usage error.
+This performs one test.
 
-In TEST_DEBUG mode, if the files do not compare, outputs file info on STDERR.
+=over 4
+
+=item ARGUMENTS
+
+A single argument is taken, considered as a hash ref, with the following keys:
+
+In TEST_DEBUG mode, if the files do not compare equal, outputs file info on
+STDERR.
+
+=over 4
+
+=item fn1
+
+B<Mandatory>  File to compare
+
+=item fn2
+
+B<Mandatory>  File to compare
+
+=item name
+
+B<Mandatory>  Test name
+
+=item sort
+
+B<Optional> sort files prior to comparison.  Requires the C<sort> command to
+be on C<$PATH> (else skips).
+
+=item gunzip
+
+B<Optional> gunzip files prior to comparison.  Requires the C<gzip> command to
+be on C<$PATH> (else skips).  gzip occurs prior to any sort.
+
+=item untar
+
+B<Optional> untar files prior to comparison.  Requires the C<tar> command to
+be on C<$PATH> (else skips).  any gzip occurs prior to any tar.  Tar files are
+considered equal if they each contain the same filenames & each file contained
+is equal.  If the sort flag is present, each file is sorted prior to comparison.
+
+=back
+
+=back
 
 =cut
 
-sub compare {
-  my ($fn1, $fn2) = @_;
+#XYZ sub _run {
+#XYZ   my ($cmd, $name, $in) = @_;
+#XYZ
+#XYZ   my $infn = defined $in ? tmpnam : '/dev/null';
+#XYZ   my $outfn = tmpnam;
+#XYZ   my $errfn = tmpnam;
+#XYZ
+#XYZ   my $pid = fork;
+#XYZ   croak "Couldn't fork: $!\n"
+#XYZ     unless defined $pid;
+#XYZ
+#XYZ   if ( $pid == 0 ) { # Child
+#XYZ     open STDOUT, '>', $outfn;
+#XYZ     open STDERR, '>', $errfn;
+#XYZ     open STDIN,  '<', $infn;
+#XYZ
+#XYZ     exec @$cmd;
+#XYZ   }
+#XYZ
+#XYZ   my $rv = waitpid $pid, 0;
+#XYZ   my $status = $?;
+#XYZ
+#XYZ   croak "Unexpected waitpid return from child $name: $rv (expected $pid)\n"
+#XYZ     unless $rv == $pid;
+#XYZ
+#XYZ   local $/ = undef;
+#XYZ   local (OUT, ERR);
+#XYZ   open *OUT, '<', $outfn;
+#XYZ   open *ERR, '<', $errfn;
+#XYZ   my $out = <OUT>;
+#XYZ   my $err = <ERR>;
+#XYZ   close *OUT;
+#XYZ   close *ERR;
+#XYZ
+#XYZ   return $status >> 8, $status & 127, $status & 128 , $out, $err
+#XYZ }
 
-  for ( grep ! defined, $fn1, $fn2 ) {
-    carp "Usage: compare($fn1, $fn2)\n"
-        if $ENV{TEST_DEBUG};
-    return -8;
-  }
+# return codes and old-style call semantics left for backwards compatibility
+BEGIN {
+  my $savewarn = $SIG{__WARN__};
+  # Subvert bizarre (& incorrect) subroutine redefined errors in 5.005_03
+  local $SIG{__WARN__} =
+    sub {
+      $savewarn->(@_)
+        if defined $savewarn                        and
+           UNIVERSAL::isa($savewarn,'CODE')         and
+           $_[0] !~ /^Subroutine compare redefined/;
+    };
 
-  {
-    my $err = 0;
+  *compare = sub {
+    my ($fn1, $fn2, $sort) = @_;
+    my ($gzip, $tar, $name);
+    my $notest = 1;
 
-    for (0..1) {
-      my $fn = ($fn1, $fn2)[$_];
-      if ( ! -e $fn ) {
-        carp "Does not exist: $fn\n"
+    if ( @_ == 1 and UNIVERSAL::isa($_[0], 'HASH') ) {
+      ($fn1, $fn2, $name, $sort, $gzip, $tar, $notest) =
+        @{$_[0]}{qw( fn1 fn2 name sort gunzip untar notest )};
+      my @missing = grep ! defined $_[0]->{$_}, qw( fn1 fn2 name );
+      carp "Missing mandatory key(s): " . join(', ', @missing) . "\n"
+        if @missing;
+    }
+
+    my ($name1, $name2) = ($fn1, $fn2);
+
+    for ( grep ! defined, $fn1, $fn2 ) {
+      carp 'Usage: compare({fn1 => $fn1, fn2 => $fn2, name => "some name"})' ."\n"
           if $ENV{TEST_DEBUG};
-        $err |= 2 ** $_;
-      } elsif ( ! -r $fn ) {
-        carp "Cannot read: $fn\n"
+      ok 0, 1, $name
+        unless $notest;
+      return -8;
+    }
+
+    {
+      my $err = 0;
+
+      for (0..1) {
+        my $fn = ($name1, $name2)[$_];
+        if ( ! -e $fn ) {
+          carp "Does not exist: $fn\n"
+            if $ENV{TEST_DEBUG};
+          $err |= 2 ** $_;
+        } elsif ( ! -r $fn ) {
+          carp "Cannot read: $fn\n"
+            if $ENV{TEST_DEBUG};
+          $err |= 2 ** $_;
+        }
+      }
+
+      if ( $err ) {
+        ok 0, 1, $name
+          unless $notest;
+        return -$err;
+      }
+    }
+
+    if ( $gzip ) {
+      unless ( find_exec('gzip') ) {
+        print "ok # Skip gzip not found in path\n";
+        return -16;
+      }
+
+      my $tmp1 = tmpnam;
+      my $tmp2 = tmpnam;
+      system "gzip $fn1 -cd > $tmp1"
+        and croak "gzip $fn1 failed: $?\n";
+      system "gzip $fn2 -cd > $tmp2"
+        and croak "gzip $fn2 failed: $?\n";
+      ($fn1, $fn2) = ($tmp1, $tmp2);
+    }
+
+    if ( $tar ) {
+      unless ( find_exec('tar') ) {
+        print "ok # Skip tar not found in path\n";
+        return -16;
+      }
+
+      local $/ = "\n";
+      chomp (my @list1 = sort qx( tar tf $fn1 ));
+      croak "tar tf $fn1 failed with wait status: $?\n"
+        if $?;
+      chomp(my @list2 = sort qx( tar tf $fn2 ));
+      croak "tar tf $fn2 failed with wait status: $?\n"
+        if $?;
+
+      if ( @list2 > @list1 ) {
+        carp
+          sprintf("More files (%d) in $name2 than $name1 (%d)\n",
+                  scalar @list2, scalar @list1)
           if $ENV{TEST_DEBUG};
-        $err |= 2 ** $_;
+        ok @list1, @list2, $name
+          unless $notest;
+        return 0;
+      } elsif ( @list1 > @list2 ) {
+        carp
+          sprintf("More files (%d) in $name1 than $name2 (%d)\n",
+                  scalar @list1, scalar @list2)
+          if $ENV{TEST_DEBUG};
+        ok @list1, @list2, $name
+          unless $notest;
+        return 0;
       }
+
+      for (my $i = 0; $i < @list1; $i++) {
+        if ( $list1[$i] lt $list2[$i] ) {
+          carp "File $list1[$i] is present in $name1 but not $name2\n"
+            if $ENV{TEST_DEBUG};
+          ok $list1[$i], $list2[$i], $name
+            unless $notest;
+          return 0;
+        } elsif ( $list1[$i] gt $list2[$i] ) {
+          carp "File $list2[$i] is present in $name2 but not $name1\n"
+            if $ENV{TEST_DEBUG};
+          ok $list2[$i], $list1[$i], $name
+            unless $notest;
+          return 0;
+        }
+      }
+
+      for my $fn (@list1) {
+        my $tmp1 = tmpnam;
+        my $tmp2 = tmpnam;
+        system "tar -xf $fn1 -O $fn > $tmp1"
+          and croak "tar -xf $fn1 -O $fn failed: $?\n";
+        system "tar -xf $fn2 -O $fn > $tmp2"
+          and croak "tar -xf $fn2 -O $fn failed: $?\n";
+        my $ok = compare({ fn1    => $tmp1,
+                           fn2    => $tmp2,
+                           sort   => $sort,
+                           notest => 1,
+                           name   =>
+                             qq'Subcheck file "$fn" for compare $name1, $name2',
+                         });
+        unless ( $ok >= 1 ) {
+          carp qq'Difference found testing file "$fn" in tars $name1 ($tmp1), $name2 ($tmp2)\n'
+            if $ENV{TEST_DEBUG};
+          ok 0, 1, $name
+            unless $notest;
+          return 0;
+        }
+      }
+
+      ok 1, 1, $name
+        unless $notest;
+      return 1;
     }
 
-    return -$err
-      if $err;
+    if ( $sort ) {
+      unless ( find_exec('sort') ) {
+        print "ok # Skip sort not found in path\n";
+        return -16;
+      }
+
+      my $tmp1 = tmpnam;
+      my $tmp2 = tmpnam;
+      system sort => $fn1, -o => $tmp1
+        and croak "Sort $fn1 failed: $?\n";
+      system sort => $fn2, -o => $tmp2
+        and croak "Sort $fn2 failed: $?\n";
+      ($fn1, $fn2) = ($tmp1, $tmp2);
+    }
+
+    unless ( File::Compare::compare($fn1, $fn2) ) {
+      ok 1, 1, $name
+        unless $notest;
+      return 1;
+    }
+
+    if ( $ENV{TEST_DEBUG} ) {
+      my $pid = fork;
+      die "Fork failed: $!\n"
+        unless defined $pid;
+
+      if ( $pid ) { # Parent
+        my $waitpid = waitpid($pid, 0);
+        die "Waitpid got: $waitpid (expected $pid)\n"
+          unless $waitpid == $pid;
+      } else { # Child
+        open *STDOUT{IO}, ">&" . fileno STDERR;
+        # Uniquify file names
+        my @args = keys %{+{ map {;$_=>1} $name1, $name2, $fn1, $fn2 }};
+        exec qw(ls -l), @args;
+      }
+
+      my $fh1 = IO::File->new($fn1, O_RDONLY)
+        or die "Couldn't open $fn1: $!\n";
+      my $fh2 = IO::File->new($fn2, O_RDONLY)
+        or die "Couldn't open $fn2: $!\n";
+
+      local $/ = "\n";
+
+      my $found = 0;
+      while ( ! $found and my $line1 = <$fh1> ) {
+        my $line2 = <$fh2>;
+        if ( ! defined $line2 ) {
+          print STDERR "$fn2 ended at line: $.\n";
+          $found = 1;
+        } elsif ( $line2 ne $line1 ) {
+          my $maxlength = max(map length($_), $line1, $line2);
+          my $minlength = min(map length($_), $line1, $line2);
+
+          my @diffchars = grep(substr($line1, $_, 1) ne substr($line2, $_, 1),
+                               0..$minlength-1);
+          my $diff = ' ' x $minlength;
+          substr($diff, $_, 1) = '|'
+            for @diffchars;
+
+          my @extrachars, map((length($line1) > length($line2) ? '^' : 'v'),
+                              $minlength..$maxlength-1);
+
+          $diff = join '', $diff, @extrachars;
+
+          my $diff_count  = @diffchars;
+          my $extra_count = @extrachars;
+
+          print STDERR <<"END";
+Difference at line $. ($diff_count characters differ) (top line is $extra_count chars longer):
+$name1:
+-->$line1<--
+   $diff
+-->$line2<--
+$name2:
+Differing characters at positions @{[join ',',@diffchars]} (zero-based)
+END
+          $found = 1;
+        }
+      }
+
+      if ( ! $found ) {
+        my $line2 = <$fh2>;
+        if ( defined $line2 ) {
+          print STDERR "$name1 ended before line: $.\n";
+        } else {
+          print STDERR "Difference between $name1, $name2 not found!\n";
+        }
+      }
+
+      close $fh1;
+      close $fh2;
+    }
+
+    ok 0, 1, $name
+      unless $notest;
+    return 0;
   }
-
-  return 1
-    unless File::Compare::compare($fn1, $fn2);
-
-  if ( $ENV{TEST_DEBUG} ) {
-    my $pid = fork;
-    die "Fork failed: $!\n"
-      unless defined $pid;
-
-    if ( $pid ) { # Parent
-      my $waitpid = waitpid($pid, 0);
-      die "Waitpid got: $waitpid (expected $pid)\n"
-        unless $waitpid == $pid;
-    } else { # Child
-      open *STDOUT{IO}, ">&" . fileno STDERR;
-      exec qw(ls -l), $fn1, $fn2;
-    }
-
-    my $fh1 = IO::File->new($fn1, O_RDONLY)
-      or die "$fn1: $!\n";
-    my $fh2 = IO::File->new($fn2, O_RDONLY)
-      or die "$fn2: $!\n";
-
-    local $/ = "\n";
-
-    my $found = 0;
-    while ( ! $found and my $line1 = <$fh1> ) {
-      my $line2 = <$fh2>;
-      if ( ! defined $line2 ) {
-        print STDERR "$fn2 ended at line: $.\n";
-        $found = 1;
-      } elsif ( $line2 ne $line1 ) {
-        print STDERR
-          "Difference at line $.:\n-1->$line1<---\n-2->$line2<---\n-";
-        $found = 1;
-      }
-    }
-
-    if ( ! $found ) {
-      my $line2 = <$fh2>;
-      if ( defined $line2 ) {
-        print STDERR "$fn1 ended before line: $.\n";
-      } else {
-        print STDERR "Difference between $fn1, $fn2 not found!\n";
-      }
-    }
-
-    close $fh1;
-    close $fh2;
-  }
-
-  return 0;
 }
 
 # -------------------------------------
@@ -755,6 +1054,98 @@ sub check_req {
   $test->(only_files([]), 1);
 }
 
+# -------------------------------------
+
+=head2 find_exec
+
+=over 4
+
+=item ARGUMENTS
+
+=over 4
+
+=item proggie
+
+The name of the program
+
+=back
+
+=item RETURNS
+
+=over 4
+
+=item path
+
+The path to the first executable file with the given name on C<$PATH>.  Or
+nothing, if no such file exists.
+
+=back
+
+=back
+
+=cut
+
+sub find_exec {
+  my ($exec) = @_;
+
+  for (split /:/, $PATH) {
+    my $try = catfile $_, $exec;
+    return $try
+      if -x $try;
+  }
+  return;
+}
+
+# -------------------------------------
+
+=head2 read_file
+
+=over 4
+
+=item ARGUMENTS
+
+=over 4
+
+=item filename
+
+B<Mandatory>
+
+=item line-terminator
+
+B<Optional>.  Value of C<$/>.  Defaults to C<"\n">.
+
+=back
+
+=item RETURNS
+
+=over 4
+
+=item lines
+
+A list of lines in the file (lines determined by the value of
+line-terminator), as an arrayref.
+
+=back
+
+=back
+
+=cut
+
+sub read_file {
+  my ($fn, $term) = @_;
+
+  $term = "\n"
+    unless defined $term;
+
+  my $fh = do { local *F; *F };
+  sysopen $fh, $fn, O_RDONLY;
+  local $/ = $term;
+  my @lines = <$fh>;
+  close $fh;
+
+  return \@lines;
+}
+
 # ----------------------------------------------------------------------------
 
 =head1 EXAMPLES
